@@ -1,5 +1,5 @@
 /* font.c -- font module.
-   Copyright (C) 2003, 2004, 2005, 2006
+   Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -1154,7 +1154,7 @@ mfont__init ()
   default_encoding.repertory_name = Mnil;
   default_encoding.repertory_charset = NULL;
   {
-    char *path, *buf;
+    char *path, *buf = NULL;
     int bufsize;
     USE_SAFE_ALLOCA;
 
@@ -1219,7 +1219,7 @@ mfont__fini ()
 MSymbol
 mfont__id (MFont *font)
 {
-  char *buf, *p;
+  char *buf = NULL, *p;
   int i;
   int file_len = (font->file == Mnil ? 0 : MSYMBOL_NAMELEN (font->file));
   int capability_len  = (font->capability == Mnil ? 0
@@ -1446,9 +1446,9 @@ mdebug_dump_font_list (MFontList *font_list)
 
   for (i = 0; i < font_list->nfonts; i++)
     {
-      fprintf (stderr, "%04X - ", font_list->fonts[i].score);
+      fprintf (mdebug__output, "%04X - ", font_list->fonts[i].score);
       mdebug_dump_font (font_list->fonts[i].font);
-      fprintf (stderr, "\n");
+      fprintf (mdebug__output, "\n");
     }
 }
 
@@ -1685,13 +1685,16 @@ mfont__get_glyph_id (MFLTFont *font, MFLTGlyphString *gstring,
 
       if (g->g.encoded)
 	continue;
-      if (! encoding->encoding_charset)
-	g->g.code = MCHAR_INVALID_CODE;
-      else if (mfont->source == MFONT_SOURCE_X && encoding->repertory_charset)
-	g->g.code = ENCODE_CHAR (encoding->repertory_charset, g->g.code);
+      if (mfont->source == MFONT_SOURCE_X && encoding->repertory_charset)
+	g->g.code = ENCODE_CHAR (encoding->repertory_charset, g->g.c);
       else
 	{
-	  unsigned code = ENCODE_CHAR (encoding->encoding_charset, g->g.code);
+	  unsigned code;
+
+	  if (encoding->encoding_charset)
+	    code = ENCODE_CHAR (encoding->encoding_charset, g->g.c);
+	  else
+	    code = g->g.code;
 
 	  if (code != MCHAR_INVALID_CODE)
 	    {
@@ -1708,9 +1711,8 @@ mfont__get_glyph_id (MFLTFont *font, MFLTGlyphString *gstring,
 			MFATAL (MERROR_FONT);
 		    }
 		}
-	      g->g.code
-		= (driver->encode_char) (rfont->frame, rfont->font, mfont,
-					 g->g.code);
+	      g->g.code = (driver->encode_char) (rfont->frame, rfont->font,
+						 mfont, code);
 	    }
 	}
       g->g.encoded = 1;
@@ -1956,7 +1958,7 @@ mfont__check_capability (MRealizedFont *rfont, MSymbol capability)
 /***ja
     @brief ȯꤹեȥץѥƥΥ.
     
-    ѿ #Mfoundry  <tt>"fonudry"</tt> 
+    ѿ #Mfoundry  <tt>"foundry"</tt> 
     Ȥ̾ĥܥǤꡢեȥץѥƥȥեץѥƥΥȤѤ롣
     ͤϡեȤγȯ̾̾ȤƻĥܥǤ롣    */
 
@@ -2395,14 +2397,14 @@ mfont_copy (MFont *font)
     If $FONT is a return value of mfont_find (), $KEY can also be one
     of the following symbols:
 
-	#Mfont_ascent, #Mfont_descent, #Mmax_advance.
+	@b Mfont_ascent, @b Mfont_descent, #Mmax_advance.
 
     @return 
     If $KEY is @c Mfoundry, @c Mfamily, @c Mweight, @c Mstyle,
     @c Mstretch, @c Madstyle, @c Mregistry, or @c Mspacing, this
     function returns the corresponding value as a symbol.  If the font
     does not have $KEY property, it returns @c Mnil.  If $KEY is @c
-    Msize, @c Mresolution, #Mfont_ascent, Mfont_descent, or
+    Msize, @c Mresolution, @b Mfont_ascent, Mfont_descent, or
     #Mmax_advance, this function returns the corresponding value as an
     integer.  If the font does not have $KEY property, it returns 0.
     If $KEY is something else, it returns @c NULL and assigns an error
@@ -3113,7 +3115,7 @@ mfont_open (MFrame *frame, MFont *font)
     @brief Encapusulate a font.
 
     The mfont_encapsulate () functions realizes a font by
-    encapusulating data $DATA or type $DATA_TYPE on #FRAME.  Currently
+    encapusulating data $DATA or type $DATA_TYPE on $FRAME.  Currently
     $DATA_TAPE is #Mfontconfig or #Mfreetype, and $DATA points to an
     object of FcPattern or FT_Face respectively.
 
@@ -3182,16 +3184,18 @@ mfont_close (MFont *font)
 /***en
     @brief Dump a font.
 
-    The mdebug_dump_font () function prints font $FONT in a human readable
-    way to the stderr.
+    The mdebug_dump_font () function prints font $FONT in a human
+    readable way to the stderr or to what specified by the environment
+    variable MDEBUG_OUTPUT_FILE.
 
     @return
     This function returns $FONT.  */
 /***ja
     @brief եȤפ.
 
-    ؿ mdebug_dump_font () ϥե $FONT  stderr 
-    ˿ʹ֤˲ɤʷǰ롣
+    ؿ mdebug_dump_font () ϥե $FONT ɸ२顼Ϥ⤷
+    Ķѿ MDEBUG_DUMP_FONT ǻꤵ줿ե˿ʹ֤˲ɤʷǽ
+    Ϥ롣
 
     @return
     δؿ $FONT ֤  */
@@ -3204,7 +3208,7 @@ mdebug_dump_font (MFont *font)
   name = xlfd_unparse_name (font, 0);
   if (name)
     {
-      fprintf (stderr, "%s", name);
+      fprintf (mdebug__output, "%s", name);
       free (name);
     }
   if (font->file != Mnil)
@@ -3216,11 +3220,11 @@ mdebug_dump_font (MFont *font)
 	if (*p == '/')
 	  lastslash = p;
       if (name)
-	fprintf (stderr, ",");
-      fprintf (stderr, "%s", lastslash + 1);
+	fprintf (mdebug__output, ",");
+      fprintf (mdebug__output, "%s", lastslash + 1);
     }
   if (font->capability != Mnil)
-    fprintf (stderr, "%s", MSYMBOL_NAME (font->capability));
+    fprintf (mdebug__output, "%s", MSYMBOL_NAME (font->capability));
   return font;
 }
 
