@@ -1,5 +1,5 @@
 /* m17n-flt.h -- header file for the FLT API of the m17n library.
-   Copyright (C) 2007
+   Copyright (C) 2007, 2008, 2009, 2010
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -109,16 +109,20 @@ typedef struct
       fractional pixel format.  */
   /***ja Ľ񤭻 26.6 fractional pixel format ɽΡ  */
   int yadv;
+  /* @{ */
   /***en Ink metrics of the glyph expressed in 26.6 fractional pixel
       format.  */
   /***ja ΥդΥ󥯥ȥå 26.6 fractional pixel format
       ɽΡ  */
   int ascent, descent, lbearing, rbearing;
+  /* @} */
+  /* @{ */
   /***en Horizontal and vertical adjustments for the glyph positioning
       expressed in 26.6 fractional pixel format.  */
   /***ja հַκݤοʿľĴͤ
       26.6 fractional pixel format ɽΡ  */
   int xoff, yoff;
+  /* @} */
   /***en Flag to tell whether the member \<code\> has already been set
       to a glyph ID in the font.  */
   /***ja С \<code\> ˴˥ ID
@@ -153,28 +157,32 @@ typedef struct
 
     The type #MFLTGlyphAdjustment is the structure to store
     information about a glyph metrics/position adjustment.  It is
-    given to the callback function #drive_otf of #MFLTFont.  */
+    given to the callback function @b drive_otf of #MFLTFont.  */
 
 /***ja
     @brief հĴΤη.
 
      #MFLTGlyphAdjustment
     ϡդΥȥå/֤Ĵ˴ؤǼ뤿ι¤ΤǤꡢ
-    #MFLTFont  callback ؿ #drive_otf Ϥ롣  */
+    #MFLTFont  callback ؿ @b drive_otf Ϥ롣  */
 
 typedef struct
 {
+  /* @{ */
   /***en Adjustments for advance width for horizontal layout and
       advance height for vertical layout expressed in 26.6 fractional
       pixel format.  */
   /***ja ʿľ̤Ĵͤ 26.6 fractional pixel format
       ɽΡ  */
   int xadv, yadv;
+  /* @} */
+  /* @{ */
   /***en Horizontal and vertical adjustments for glyph positioning
       expressed in 26.6 fractional pixel format.  */
   /***ja հַ᤿οʿľĴͤ 26.6 fractional pixel
       format ɽΡ  */
   int xoff, yoff;
+  /* @} */
   /***en Number of glyphs to go back for drawing a glyph.  */
   /***ja Τ٤տ  */
   short back;
@@ -233,37 +241,65 @@ typedef struct
 
     The type #MFLTOtfSpec is the structure that contains information
     about the GSUB and GPOS features of a specific script and language
-    system to be applied to a glyph sequence.  */
+    system.  The information is used to select which features to
+    apply to a glyph string, or to check if a specific FLT is usable
+    for a specific font.  */
 
 /***ja
     @brief GSUB  GPOS OpenType ơ֥λͤΤη.
 
      #MFLTOtfSpec ϡGSUB  GPOSե㡼ξǼ뤿
     ι¤ΤǤ롣ե㡼ΥץȤӸ쥷
-    ƥΤΤǤꡢŬѤ롣  */
+    ƥΤΤǤ롣ξϡɤΥե㡼򥰥ŬѤ
+    뤫뤤 FLT ΥեȤФͭɤη
+    ˻Ѥ롣  */
 
 typedef struct
 {
   /***en Unique symbol representing the spec.  This is the same as the
-      #OTF-SPEC of the FLT.  */
+      @ref OTF-SPEC of the FLT.  */
   /***ja λͤɽ魯ˡʥܥ롣
-      FLT  #OTF-SPEC ƱͤǤ롣  */
+      FLT  @ref OTF-SPEC ƱͤǤ롣  */
   MSymbol sym;
 
+  /* @{ */
   /***en Tags for script and language system.  */
   /***ja ץȤӸ쥷ƥΥ  */
   unsigned int script, langsys;
+  /* @} */
 
-  /***en Array of GSUB (1st element) and GPOS (2nd element) features.
-      Each array is terminated by 0.  If an element is 0xFFFFFFFF
-      apply the previous features in that order, and apply all the
-      other features except those that appear in the following elements.
-      It may be NULL if there are no features.  */
-  /***ja GSUB ե㡼1ǡGPOS ե㡼2ǤȤ
-      󡣳0Ǽ롣⤷Ǥ 0xFFFFFFFFʤС
-      ե㡼򤽤νŬѤ˰ʹߤǤȤƸ
-      ե㡼ʳΤ٤ƤŬѤ롣ե㡼1Ĥʤ
-       NULL Ǥ褤  */
+  /***en Array of GSUB (1st element) and GPOS (2nd element) feature
+      tag arrays.  Each array is terminated by 0.  It may be NULL if
+      there is no feature to specify.
+
+      (1) The case of using this information for selecting which
+      features to apply to a glyph string.  If the array is NULL,
+      apply no feature.  If the first element is 0xFFFFFFFF, apply all
+      available features except for what appear in the second and
+      following elements (if any).  Otherwise, apply all listed
+      features.
+
+      (2) The case of using this information for checking if a a font
+      can be drived by a specific FLT.  If the array is NULL, the font
+      should not have any features.  Otherwize, the font should have
+      all features before 0xFFFFFFFF element (if any) and should not
+      have any features after that element.  */
+  /***ja GSUB ե㡼1ǡGPOS ե㡼
+      2ǤȤ󡣳0Ǽ롣ե㡼
+      λ꤬1ĤʤϤǤ NULL Ǥ褤
+
+      (1) ξ󤬥ŬѤ٤ե㡼˻Ȥ
+      硣⤷󼫿ȤNULLʤ顢ɤΥե㡼ŬѤʤ⤷
+      Ǥ 0xFFFFFFFF ʤ顢ܰʹߤΥե㡼ʤ⤷
+      Сˤ٤ƤŬѲǽʥե㡼ŬѤ롣ʳξ
+      ꥹȤ줿٤ƤΥե㡼ŬѤ롣
+
+      (2) ξ FLT ΥեȤͭɤη˻
+      硣⤷󼫿ȤNULLʤ顢եȤϥե㡼
+      äƤƤϤʤ⤷ǽǤ0xFFFFFFFFʤ顢ե
+      ϣܤǰʹߤΥեȤäƤƤϤʤʳξ
+      硢եȤ0xFFFFFFFF Τ٤ƤΥե㡼
+      0xFFFFFFFF ʹߤΥե㡼ϰĤäƤƤϤʤ*/
   unsigned int *features[2];
 } MFLTOtfSpec;
 
@@ -273,13 +309,22 @@ typedef struct
     @brief Type of font to be used by the FLT driver.
 
     The type #MFLTFont is the structure that contains information
-    about a font used by the FLT driver.  */
+    about a font used by the FLT driver.  Usually, an application
+    should prepare a bigger structure whose first element is MFLTFont
+    and has more information about the font that is used by callback
+    funcitons, and give that structure to mflt functions by coercing
+    it to MFLTFont.  It is assured that callback functions can safely
+    coerce MFLTFont back to the original structure.  */
 
 /***ja
     @brief FLT ɥ饤ФȤեȤη.
 
      #MFLTFont ϡFLTɥ饤ФȤեȤ˴ؤǼ뤿
-    ι¤ΤǤ롣  */
+    ι¤ΤǤ롣̾異ץꥱϺǽǤ MFLTFont ǡ
+    ĤǤcallbackؿѤեȾä礭
+    ¤ΤѰդ MFLTFont  coerce  mflt γƴؿϤ
+    callbackؿ MFLTFont 򸵤ι¤Τ coerce ľȤǤ
+    ȤݾڤƤ롣 */
 
 typedef struct _MFLTFont
 {
@@ -291,9 +336,12 @@ typedef struct _MFLTFont
        (Ȥ OpenTypeեȤξʤ) ϡ#Mnil Ǥ褤 */
   MSymbol family;
 
-  /***en Horizontal and vertical font sizes in pixels per EM.  */
-  /***ja եȤοʿľ pixels per EM ɽΡ  */
-  int x_ppem, y_ppem;
+  /***en Horizontal font sizes in pixels per EM.  */
+  /***ja եȤοʿ pixels per EM ɽΡ  */
+  int x_ppem;
+  /***en Vertical font sizes in pixels per EM.  */
+  /***ja եȤοľ pixels per EM ɽΡ  */
+  int y_ppem;
 
   /***en Callback function to get glyph IDs for glyphs between FROM
      (inclusive) and TO (exclusive) of GSTRING.  If the member \<encoded\>
@@ -377,6 +425,18 @@ extern MCharTable *mflt_coverage (MFLT *flt);
 extern int mflt_run (MFLTGlyphString *gstring, int from, int to,
 		     MFLTFont *font, MFLT *flt);
 
+extern int mflt_enable_new_feature;
+
+extern MSymbol (*mflt_font_id) (MFLTFont *font);
+
+extern int (*mflt_iterate_otf_feature) (MFLTFont *font,
+					MFLTOtfSpec *spec,
+					int from, int to,
+					unsigned char *table);
+
+extern int (*mflt_try_otf) (struct _MFLTFont *font, MFLTOtfSpec *spec,
+			    MFLTGlyphString *gstring, int from, int to);
+
 /*=*/
 /*** @} */
 
