@@ -1,5 +1,5 @@
 /* m17n.h -- header file for the SHELL API of the m17n library.
-   Copyright (C) 2003, 2004
+   Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
      National Institute of Advanced Industrial Science and Technology (AIST)
      Registration Number H15PRO112
 
@@ -853,22 +853,28 @@ extern MText *mtext_toupper (MText *mt);
 
 /*** @addtogroup m17nInputMethod 
      @{ */
-
+/*=*/
 /* Struct forward declaration.  */
+/***
+    @brief See struct MInputMethod */
 typedef struct MInputMethod MInputMethod;
+/*=*/
+/***
+    @brief See struct MInputContext */
 typedef struct MInputContext MInputContext;
+/*=*/
 
 /***en
     @brief Type of input method callback functions.
 
     This is the type of callback functions called from input method
-    drivers.  #IC is a pointer to an input context, #COMMAND is a name
+    drivers.  $IC is a pointer to an input context, $COMMAND is a name
     of callback for which the function is called.   */
 /***ja
     @brief ϥ᥽åɥХåؿη.
 
-    ϥ᥽åɤƤФ륳ХåؿηǤ롣#IC 
-    ϥƥȤؤΥݥ󥿡#COMMAND ϴؿƤФ륳Хå̾Ǥ롣  */
+    ϥ᥽åɤƤФ륳ХåؿηǤ롣$IC 
+    ϥƥȤؤΥݥ󥿡$COMMAND ϴؿƤФ륳Хå̾Ǥ롣  */
 
 typedef void (*MInputCallbackFunc) (MInputContext *ic, MSymbol command);
 /*=*/
@@ -1039,31 +1045,34 @@ typedef struct MInputDriver
       @brief List of callback functions.
 
       List of callback functions.  Keys are one of
-      #Minput_preedit_start, #Minput_preedit_draw,
-      #Minput_preedit_done, #Minput_status_start, #Minput_status_draw,
-      #Minput_status_done, #Minput_candidates_start,
-      #Minput_candidates_draw, #Minput_candidates_done,
-      #Minput_set_spot, #Minput_toggle, #Minput_reset,
-      #Minput_get_surrounding_text, #Minput_delete_surrounding_text.
+      @b Minput_preedit_start, @b Minput_preedit_draw,
+      @b Minput_preedit_done, @b Minput_status_start, @b Minput_status_draw,
+      @b Minput_status_done, @b Minput_candidates_start,
+      @b Minput_candidates_draw, @b Minput_candidates_done,
+      @b Minput_set_spot, @b Minput_toggle, @b Minput_reset,
+      @b Minput_get_surrounding_text, @b Minput_delete_surrounding_text.
       Values are functions of type #MInputCallbackFunc.  */
   /***ja
       @brief ХåؿΥꥹ.
 
       ХåؿΥꥹȡϼΤ줫
-      #Minput_preedit_start, #Minput_preedit_draw,
-      #Minput_preedit_done, #Minput_status_start, #Minput_status_draw,
-      #Minput_status_done, #Minput_candidates_start,
-      #Minput_candidates_draw, #Minput_candidates_done,
-      #Minput_set_spot, #Minput_toggle, #Minput_reset,
-      #Minput_get_surrounding_text, #Minput_delete_surrounding_text
+      @b Minput_preedit_start, @b Minput_preedit_draw,
+      @b Minput_preedit_done, @b Minput_status_start, @b Minput_status_draw,
+      @b Minput_status_done, @b Minput_candidates_start,
+      @b Minput_candidates_draw, @b Minput_candidates_done,
+      @b Minput_set_spot, @b Minput_toggle, @b Minput_reset,
+      @b Minput_get_surrounding_text, @b Minput_delete_surrounding_text
       ͤ#MInputCallbackFunc δؿ  */
   MPlist *callback_list;
 
 } MInputDriver;
 /*=*/
+/*** @} */
+/*=*/
 
 extern MInputDriver minput_default_driver;
 
+extern MSymbol Minput_method;
 extern MSymbol Minput_driver;
 
 extern MInputDriver *minput_driver;
@@ -1094,6 +1103,9 @@ extern MSymbol Minherited;
 extern MSymbol Mcustomized;
 extern MSymbol Mconfigured;
 
+/*** @addtogroup m17nInputMethod 
+     @{ */
+/*=*/
 /***en
     @brief Structure of input method.
 
@@ -1177,7 +1189,7 @@ struct MInputContext
       ˤäꤵ롣  */
   MText *produced;
 
-  /***en Argument given to the function minput_create_im (). */
+  /***en Argument given to the function minput_create_ic (). */
   /***ja ؿ minput_create_ic () Ϥ */
   void *arg;
 
@@ -1216,15 +1228,15 @@ struct MInputContext
   /***en The usage of the following members depends on the input
       method driver.  The descriptions below are for the driver of an
       internal input method.  They are set by the function
-      <im>->driver.filter ().  */
+      \<im\>->driver.filter ().  */
   /***ja ʲΥФλˡϥ᥽åɥɥ饤Фˤäưۤʤ롣
       ʲϡϥ᥽åѤϥɥ饤ФФΤǤ롣
-      ϴؿ <im>->driver.filter () ˤäꤵ롣  */
+      ϴؿ \<im\>->driver.filter () ˤäꤵ롣  */
 
-  /***en Pointer to extra information that <im>->driver.create_ic ()
+  /***en Pointer to extra information that \<im\>->driver.create_ic ()
       setups.  It is used to record the internal state of the input
       context.  */
-  /***ja <im>->driver.create_ic () ꤹɲþؤΥݥ󥿡
+  /***ja \<im\>->driver.create_ic () ꤹɲþؤΥݥ󥿡
       ϥƥȤ֤Ͽ뤿Ѥ롣 */
   void *info;
 
@@ -1233,21 +1245,21 @@ struct MInputContext
   /***ja ϥƥȤθߤξ֤ɽ M-text  */
   MText *status;
 
-  /***en The function <im>->driver.filter () sets the value to 1 when
+  /***en The function \<im\>->driver.filter () sets the value to 1 when
       it changes \<status\>.  */
-  /***ja ؿ <im>->driver.filter () ϡ\<status\> Ѥݤˤͤ 1 
+  /***ja ؿ \<im\>->driver.filter () ϡ\<status\> Ѥݤˤͤ 1 
       ꤹ롣  */
   int status_changed;
 
   /***en M-text containing the current preedit text.  The function
-      <im>->driver.filter () sets the value.  */
+      \<im\>->driver.filter () sets the value.  */
   /***ja ߤ preedit ƥȤޤ M-textؿ 
-      <im>->driver.filter () ˤäꤵ롣  */ 
+      \<im\>->driver.filter () ˤäꤵ롣  */ 
   MText *preedit;
 
-  /***en The function <im>->driver.filter () sets the value to 1 when
+  /***en The function \<im\>->driver.filter () sets the value to 1 when
       it changes \<preedit\>.  */
-  /***ja ؿ <im>->driver.filter () ϡ\<preedit\> Ѥݤˤͤ 
+  /***ja ؿ \<im\>->driver.filter () ϡ\<preedit\> Ѥݤˤͤ 
       1 ꤹ롣  */
   int preedit_changed;
 
@@ -1255,9 +1267,9 @@ struct MInputContext
   /***ja \<preedit\>Υ  */
   int cursor_pos;
 
-  /***en The function <im>->driver.filter () sets the value to 1 when
+  /***en The function \<im\>->driver.filter () sets the value to 1 when
       it changes \<cursor_pos\>.  */
-  /***ja ؿ <im>->driver.filter () ϡ\<cursor_pos\> Ѥݤˤͤ 
+  /***ja ؿ \<im\>->driver.filter () ϡ\<cursor_pos\> Ѥݤˤͤ 
       1 ꤹ롣  */
   int cursor_pos_changed;
 
@@ -1283,45 +1295,49 @@ struct MInputContext
       */
   int candidate_index;
 
+  /* @{ */
   /***en Start and the end positions of the preedit text where
        \<candidate_list\> corresponds to.  */
   /***ja preedit ƥǡ\<candidate_list\>бǽȺǸΰ֡
        */
   int candidate_from, candidate_to;
+  /* @} */
 
   /***en Flag telling whether the current candidate group must be
-      shown or not.  The function <im>->driver.filter () sets the
+      shown or not.  The function \<im\>->driver.filter () sets the
       value to 1 when an input method required to show candidates, and
       sets the value to 0 otherwise.  */
   /***ja ߤθ䥰롼פɽ뤫ɤ򼨤ե饰
-      ؿ <im>->driver.filter () ϡϥ᥽åɤɽ׵ᤷͤ
+      ؿ \<im\>->driver.filter () ϡϥ᥽åɤɽ׵ᤷͤ
       1 ˡʳλ 0 ꤹ롣  */
   int candidate_show;
 
-  /***en The function <im>->driver.filter () sets the value to bitwise
+  /***en The function \<im\>->driver.filter () sets the value to bitwise
       OR of @c enum @c MInputCandidatesChanged when it changed any of
       the above members (\<candidate_XXX\>), and sets the value to 0
       otherwise.  */
-  /***ja ؿ <im>->driver.filter () ϡ嵭Υ \<candidate_XXX\>
+  /***ja ؿ \<im\>->driver.filter () ϡ嵭Υ \<candidate_XXX\>
       ΣĤǤѹݤˤϡͤ @c enum @c
       MInputCandidatesChanged Υӥåñ̤Ǥ OR ꤹ롣
       Ǥʤ 0 ꤹ롣 */
   int candidates_changed;
 
-  /***en Plist that can be freely used by <im>->driver functions.
+  /***en Plist that can be freely used by \<im\>->driver functions.
       The driver of internal input method uses it to exchange extra
       arguments and result for callback functions.  The function
-      <im>->driver.create_ic () sets this to an empty plist, and the
-      function <im>->driver.destroy_ic () frees it by using
+      \<im\>->driver.create_ic () sets this to an empty plist, and the
+      function \<im\>->driver.destroy_ic () frees it by using
       m17n_object_unref ().  */
-  /***ja <im>->driver δؿˤäƼͳ˻ѤǤ plist
+  /***ja \<im\>->driver δؿˤäƼͳ˻ѤǤ plist
       ϥ᥽åѥɥ饤ФϤ򥳡ХåؿȤΰ
-      μϤ˻Ѥ롣ؿ <im>->driver.create_ic () Ϥ plist
-      ꤹ롣ؿ<im>->driver.destroy_ic () 
+      μϤ˻Ѥ롣ؿ \<im\>->driver.create_ic () Ϥ plist
+      ꤹ롣ؿ\<im\>->driver.destroy_ic () 
       m17n_object_unref () ѤƤ plist 롣  */
   MPlist *plist;
 };
 
+/*=*/
+/*** @} */
 /*=*/
 
 extern MInputMethod *minput_open_im (MSymbol language, MSymbol name,
@@ -1372,12 +1388,9 @@ extern MPlist *minput_get_variables (MSymbol language, MSymbol name);
 
 extern int minput_set_variable (MSymbol language, MSymbol name,
 				MSymbol variable, void *value);
-/*=*/
-/***     @} */
 
 extern MInputMethod *mdebug_dump_im (MInputMethod *im, int indent);
 
-
 M17N_END_HEADER
 
 #endif /* _M17N_H_ */
